﻿/*
    Universität Hamburg
	Fakultät für Mathematik, Informatik und Naturwissenschaften
	Fachbereich Informatik
	Arbeitsbereich Wissenschaftliches Rechen
	Seminar "Effiziente Programmierung in C"
	Thema "CPU-Caches"
	Beispiel "Matrizenmultiplikation"
    Copyright (C) 2014  Christian Dusse

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

/**
 * Die Größe der quadratischen Arrays
 */
#define SIZE 1000

/**
 * Die zu messende Funktion
 */
void foo(double ** p1, double ** p2, double ** p3)
{
#ifndef OPTIMIZED
    for (int i = 0; i < SIZE; i++)
    {
        for (int j = 0; j < SIZE; j++)
        {
            for (int k = 0; k < SIZE; k++)
            {
                p3[i][j] += p1[i][k] * p2[k][j];
            }
        }
    }
#else
    // Alloziieren des Speichers für das temporäre Array
    double ** temp = malloc(SIZE * sizeof(double*));
    for (int i = 0; i < SIZE; i++)
    {
	    // Alloziieren des Speichers für die Subarrays des temporären Arrays
        temp[i] = malloc(SIZE * sizeof(double));
    }
	// Transponieren der Matrix
    for (int i = 0; i < SIZE; i++)
    {
        for (int j = 0; j < SIZE; j++)
        {
            temp[i][j] = p2[j][i];
        }
    }
    for (int i = 0; i < SIZE; i++)
    {
        for (int j = 0; j < SIZE; j++)
        {
            for (int k = 0; k < SIZE; k++)
            {
                p3[i][j] += p1[i][k] * temp[j][k];
            }
        }
    }
    // Freigeben des Speichers und Zurücksetzen der Pointer
	for (int i = 0; i < SIZE; i++)
    {
        free(temp[i]);
		temp[i] = NULL;
    }
    free(temp);
	temp = NULL;
#endif
}

int main(int argc, char *argv[])
{
    // Alloziieren des Speichers für das erste Array
	double ** p1 = malloc(SIZE * sizeof(double*));
	if(p1 == NULL)
	{
	    // Bei Scheitern zumindest eine kurze Fehlermeldung
	    printf("%s", "error");
	}
	// Alloziieren des Speichers für die Subarrays des ersten Arrays
    for (int i = 0; i < SIZE; i++)
    {
        p1[i] = malloc(SIZE * sizeof(double));
		if(p1[i] == NULL)
		{
		    // Bei Scheitern zumindest eine kurze Fehlermeldung
	        printf("%s", "error");
		}
    }

	// Alloziieren des Speichers für das zweite Array
    double ** p2 = malloc(SIZE * sizeof(double*));
	if(p2 == NULL)
	{
	    // Bei Scheitern zumindest eine kurze Fehlermeldung
	    printf("%s", "error");
	}
	// Alloziieren des Speichers für die Subarrays des zweiten Arrays
    for (int i = 0; i < SIZE; i++)
    {
        p2[i] = malloc(SIZE * sizeof(double));
		if(p2[i] == NULL)
		{
		    // Bei Scheitern zumindest eine kurze Fehlermeldung
	        printf("%s", "error");
		}
    }

	// Alloziieren des Speichers für das dritte Array
    double ** p3 = malloc(SIZE * sizeof(double*));
	if(p3 == NULL)
	{
	    // Bei Scheitern zumindest eine kurze Fehlermeldung
	    printf("%s", "error");
	}
	// Alloziieren des Speichers für die Subarrays des dritten Arrays
    for (int i = 0; i < SIZE; i++)
    {
        p3[i] = malloc(SIZE * sizeof(double));
		if(p3[i] == NULL)
		{
		    // Bei Scheitern zumindest eine kurze Fehlermeldung
	        printf("%s", "error");
		}
    }
	
	// Start der Zeitmessung
	double tstart = (double) clock() / CLOCKS_PER_SEC;

	// Aufruf der zu messenden Funktion
    foo(p1,p2,p3);
	
	// Ende der Zeitmessung
	double tstop = (double) clock() / CLOCKS_PER_SEC;
	
	// Freigeben des Speichers und Zurücksetzen der Pointer
	for (int i = 0; i < SIZE; i++)
    {
        free(p1[i]);
		p1[i] = NULL;
    }
    free(p1);
	p1 = NULL;
	
	for (int i = 0; i < SIZE; i++)
    {
        free(p2[i]);
		p2[i] = NULL;
    }
    free(p2);
	p2 = NULL;
	
	for (int i = 0; i < SIZE; i++)
    {
        free(p3[i]);
		p3[i] = NULL;
    }
    free(p3);
    p3 = NULL;

	// Ausgabe der Zeitmessung
    printf("%f", tstop - tstart);
	
    return 0;
}
